/* File: filesync.c */


#include "filesync_functions.h"

int sock, queue_size, array_size, tree_height, tree_fanout, running = 1;
char *directory;
qhead_ptr server_queue, threads_queue, calc_queue;
pthread_t *threads;
pthread_mutex_t queue_empty_mtx = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t queue_empty_cond = PTHREAD_COND_INITIALIZER;


int main(int argc, char** argv)
{
    int i;
    struct sockaddr_in servername;
    struct in_addr inp;
    array_size = ((((int) pow(tree_fanout, tree_height + 1)) - 1) / (tree_fanout - 1));    
    /*an erthei sima SIGPIPE to agnoei kai diaxeirizetai to sfalma stis read kai write*/
    signal(SIGPIPE, SIG_IGN);

    /*elengxos tou arithmou ton arguments*/
    if (argc < 8)
    {
        printf("Not enough arguments! The program will now exit...\n");
        return 0;
    }

    /*perasma tous se metavlites*/
    char *serverhost = argv[1];
    int result = inet_aton(serverhost, &inp);
    if (!result)
    {
        printf("Invalid IP address! Try again..\n");
        return 0;
    }
    int port = atoi(argv[2]);
    if (port < 0 || port > 65535)
    {
        printf("Port number must be within 0-65535 range! Try again..\n");
        return 0;
    }
    int pool_size = atoi(argv[3]);
    queue_size = atoi(argv[4]);
    tree_height = atoi(argv[5]);
    tree_fanout = atoi(argv[6]);
    directory = argv[7];
    if (!directory_exists(directory))
    {
        return 0;
    }

    /*Dimiourgia ouras minimaton apo server*/
    if (queue_init(&server_queue))
    {       
        return 1;
    }
    /*Dimiourgia ouras minimaton gia epeksergasia apo ta threads*/
    if (queue_init(&calc_queue))
    {
        queue_destroy(server_queue, array_size);        
        return 1;
    }
    /*Dimiourgia ouras minimaton ton apotelesmaton ton threads*/
    if (queue_init(&threads_queue))
    {
        queue_destroy(server_queue, array_size);    
        queue_destroy(calc_queue, array_size);        
        return 1;
    }          

    /*Dimiourgia tou threads pool*/
    if ((threads = malloc(pool_size * sizeof (pthread_t))) == NULL)
    {
        perror("malloc failed");
        queue_destroy(server_queue, array_size);
        queue_destroy(threads_queue, array_size);
        queue_destroy(calc_queue, array_size);        
        return 1;
    }
    for (i = 0; i < pool_size; i++)
    {
        if (pthread_create(&threads[i], NULL, worker_thread, NULL))
        {
            perror("pthread_create failed");
            queue_destroy(server_queue, array_size);
            queue_destroy(threads_queue, array_size);
            queue_destroy(calc_queue, array_size);            
            free(threads);
            return 1;
        }

    }
    
    create_messages();
    
    /*Dimiourgei to socket gia tin epikoinonia me ton server*/
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0)
    {
        perror("socket failed");
        queue_destroy(server_queue, array_size);
        queue_destroy(threads_queue, array_size);
        queue_destroy(calc_queue, array_size);            
        free(threads);        
        return 1;
    }
    
    int opt = 1;
    /*Apenergopoiisi tou Nagle buffering algorithmou*/
    if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *) &opt, sizeof (opt)) < 0)
    {
        perror("setsockopt failed");
        queue_destroy(server_queue, array_size);
        queue_destroy(threads_queue, array_size);
        queue_destroy(calc_queue, array_size);
        free(threads);
        close(sock);
        return 1;
    }    

    memset(&servername, 0, sizeof (servername));
    servername.sin_family = AF_INET;
    servername.sin_addr = inp;
    servername.sin_port = htons(port);

    /*Syndeetai ston server*/
    if (connect(sock, (struct sockaddr *) &servername, sizeof (servername)) < 0)
    {
        perror("connect failed");
        queue_destroy(server_queue, array_size);
        queue_destroy(threads_queue, array_size);
        queue_destroy(calc_queue, array_size);
        free(threads);
        close(sock);      
        return 1;
    }    
    
    /*Stelnei ta dedomena ston server*/
    server_getfiles(sock);
    /*Lamvanei tin apantisi apo ton server*/
    server_response(sock);
    
    pthread_mutex_lock(&queue_empty_mtx);
    while (queue_not_empty(calc_queue) && running)
    {
        pthread_cond_wait(&queue_empty_cond, &queue_empty_mtx);
    }
    pthread_mutex_unlock(&queue_empty_mtx);
    
    running = 0;
    
    for(i = 0; i < pool_size; i++)
    {
        pthread_mutex_lock(&queue_empty_mtx);
        pthread_cond_signal(&queue_empty_cond);
        pthread_mutex_unlock(&queue_empty_mtx);        
    }
    
    for(i = 0; i < pool_size; i++)
    {
        pthread_join(threads[i], NULL);
    }
    
    printf("Threads Queue:\n");
    queue_print(threads_queue);

    synchronize();
        
    queue_destroy(server_queue, array_size);
    queue_destroy(threads_queue, array_size);
    queue_destroy(calc_queue, array_size);
    free(threads);
    close(sock);
    
    return 0;
}
