/*File: filesync_functions.c*/


#include "filesync_functions.h"


/*mutex kai condition gia gemati oura*/
pthread_mutex_t queue_full_mtx = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t queue_full_cond = PTHREAD_COND_INITIALIZER;
/*mutex gia tin eisaggogi stin oura threads_queue*/
pthread_mutex_t tqueue_ins_mtx = PTHREAD_MUTEX_INITIALIZER;

int directory_exists(char *directory)
{
    struct stat s;
    int err = stat(directory, &s);
    /*an epistrepsei sfalma typonei ton typo tou sfalmatos*/
    if (err == -1)
    {
        if (ENOENT == errno)
        {
            printf("\nGiven directory does not exist! Try again...\n\n");
        }
        else
        {
            perror("stat");
        }
    }
    /*allios tsekarei an teleionei me ton xaraktira '/' kai an einai directory*/
    else
    {
        if (S_ISDIR(s.st_mode))
        {
            int len = strlen(directory);
            if (directory[len - 1] != '/')
            {
                printf("\nDirectory name must end with the '/' character! Try again...\n\n");
                return 0;
            }
            return 1;
        }
        else
        {
            printf("\nGiven directory is not a directory! Try again...\n\n");
        }
    }

    return 0;
}

void create_messages()
{
    DIR *dir_ptr;
    struct dirent *dirent_ptr;

    if ((dir_ptr = opendir(directory)) == NULL)
    {
        perror("opendir failed");
        queue_destroy(server_queue, array_size);
        queue_destroy(threads_queue, array_size);
        queue_destroy(calc_queue, array_size);
        free(threads);
        exit(EXIT_FAILURE);
    }
    else
    {
        /*vriskei ta diathesima arxeia tou fakelou*/
        while ((dirent_ptr = readdir(dir_ptr)) != NULL)
        {
            if (strcmp(dirent_ptr->d_name, ".") && strcmp(dirent_ptr->d_name, ".."))
            {
                /*gia kathe ena apo auta dimiourgei ena processfile minima*/
                qnode_ptr temp_node = malloc(sizeof (qnode));
                temp_node->msg.id = CALCULATE_HASH;
                temp_node->msg.filename = malloc(strlen(dirent_ptr->d_name) + 1);
                strcpy(temp_node->msg.filename, dirent_ptr->d_name);

                pthread_mutex_lock(&queue_empty_mtx);
                while (calc_queue->size == queue_size && running)
                {
                    pthread_mutex_unlock(&queue_empty_mtx);
                    pthread_mutex_lock(&queue_full_mtx);
                    pthread_cond_wait(&queue_full_cond, &queue_full_mtx);
                    pthread_mutex_unlock(&queue_full_mtx);
                    pthread_mutex_lock(&queue_empty_mtx);
                }

                /*to eisagei stin oura*/
                queue_insert(calc_queue, temp_node);
                pthread_mutex_unlock(&queue_empty_mtx);
                pthread_cond_signal(&queue_empty_cond);

                fflush(stdout);
            }
        }
        closedir(dir_ptr);
    }
}

void server_getfiles()
{
    int nbytes, dir_size;
    char tmp[12];

    message *msg = malloc(sizeof (message));
    msg->id = GETFILES;
    
    /*stelnei ston server to minima getfiles*/
    nbytes = write(sock, msg, sizeof (message));
    if (nbytes < 0)
    {
        perror("write");
        free(msg);
        queue_destroy(server_queue, array_size);
        queue_destroy(threads_queue, array_size);
        queue_destroy(calc_queue, array_size);
        free(threads);
        close(sock);
        exit(EXIT_FAILURE);
    }

    /*stelnei to megethos tou onomatos tou directory*/
    dir_size = strlen(directory) + 1;
    sprintf(tmp, "%d", dir_size);
    nbytes = write(sock, tmp, 12 * sizeof (char));
    if (nbytes < 0)
    {
        perror("write");
        free(msg);
        queue_destroy(server_queue, array_size);
        queue_destroy(threads_queue, array_size);
        queue_destroy(calc_queue, array_size);
        free(threads);
        close(sock);
        exit(EXIT_FAILURE);
    }

    /*stelnei to onoma tou directory*/
    nbytes = write(sock, directory, dir_size);
    if (nbytes < 0)
    {
        perror("write");
        free(msg);
        queue_destroy(server_queue, array_size);
        queue_destroy(threads_queue, array_size);
        queue_destroy(calc_queue, array_size);
        free(threads);
        close(sock);
        exit(EXIT_FAILURE);
    }

    free(msg);
}

void server_response()
{
    int nbytes, num = -1, i = 0;

    while (num != i)
    {
        qnode_ptr queue_msg = malloc(sizeof (qnode));
        
        /*lamvanei to minima apo ton server*/
        nbytes = read(sock, &(queue_msg->msg), sizeof (message));
        if (nbytes < 0)
        {
            perror("read");
            free(queue_msg);
            queue_destroy(server_queue, array_size);
            queue_destroy(threads_queue, array_size);
            queue_destroy(calc_queue, array_size);
            free(threads);
            close(sock);
            exit(EXIT_FAILURE);
        }

        /*analoga me ton typo tou minimatos ektelei tis aparaitites energeies*/
        if (queue_msg->msg.id == FILEINFO)
        {
            char size_buf[12];
            int fname_size;

            /*lamvanei to mikos tou onomatos tou arxeiou*/
            nbytes = read(sock, size_buf, 12 * sizeof (char));
            if (nbytes < 0)
            {
                perror("read");
                free(queue_msg);
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }
            fname_size = atoi(size_buf);

            /*lamvanei to onoma tou arxeiou*/
            queue_msg->msg.filename = malloc(fname_size);
            nbytes = read(sock, queue_msg->msg.filename, fname_size);
            if (nbytes < 0)
            {
                perror("read");
                free(queue_msg);
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }

            /*etoimazei to dentro pou tha mpei i pliroforia gia to hash tree tou arxeiou*/
            int j;
            int array_size = ((((int) pow(tree_fanout, tree_height + 1)) - 1) / (tree_fanout - 1));
            if ((queue_msg->msg.hash_tree = malloc(array_size * sizeof (unsigned int*))) == NULL)
            {
                perror("malloc failed");
                free(queue_msg);
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }
            for (j = 0; j < array_size; j++)
            {
                if ((queue_msg->msg.hash_tree[j] = malloc(4 * sizeof (unsigned int))) == NULL)
                {
                    perror("malloc failed");
                    free(queue_msg->msg.hash_tree);
                    free(queue_msg);
                    queue_destroy(server_queue, array_size);
                    queue_destroy(threads_queue, array_size);
                    queue_destroy(calc_queue, array_size);
                    free(threads);
                    close(sock);
                    exit(EXIT_FAILURE);
                }
            }

            int nodes_num = ((((int) pow(tree_fanout, tree_height)) - 1) / (tree_fanout - 1));
            int payload = queue_msg->msg.filesize / PAGE_SIZE;
            if (queue_msg->msg.filesize % PAGE_SIZE != 0)
            {
                payload += 1;
            }
            int tree_size = (nodes_num + payload) * 4 * sizeof (unsigned int);
            char *tree_buf = malloc(tree_size);

            int total_bytes = 0;
            nbytes = 0;
            /*lamvanei kai apothkeuei se ena buffer tin pliroforia gia to hash tree*/
            while (total_bytes < tree_size)
            {
                nbytes = read(sock, tree_buf + total_bytes, tree_size - total_bytes);
                if (nbytes < 0)
                {
                    perror("read");
                    free(queue_msg->msg.hash_tree);
                    free(queue_msg);
                    queue_destroy(server_queue, array_size);
                    queue_destroy(threads_queue, array_size);
                    queue_destroy(calc_queue, array_size);
                    free(threads);
                    close(sock);
                    exit(EXIT_FAILURE);
                }
                total_bytes += nbytes;
            }

            /*pernaei tin pliroforia apo ton buffer sto dentro*/
            for (j = 0; j < nodes_num + payload; j++)
            {
                memcpy(queue_msg->msg.hash_tree[j], tree_buf + j * 4 * sizeof (unsigned int), 4 * sizeof (unsigned int));
            }

            /*gemizei tis ypoloipes kenes selides me 0*/
            for (; j < array_size; j++)
            {
                memset(queue_msg->msg.hash_tree[j], 0, 4 * sizeof (unsigned int));
            }

            free(tree_buf);

            queue_insert(server_queue, queue_msg);
            i++;
        }
        else if (queue_msg->msg.id == EXPECT)
        {
            /*apothikeuei ton arithmo me ta arxeia pou tha perimenei na erthoun*/
            num = queue_msg->msg.expect_num;
            free(queue_msg);
            printf("Client expects %d files!\n", num);
        }
        else if (queue_msg->msg.id == ERROR)
        {
            char size_buf[12];
            int error_size;

            /*lamvanei to megethos tou onomatos tou sfalmatos*/
            nbytes = read(sock, size_buf, 12 * sizeof (char));
            if (nbytes < 0)
            {
                perror("read");
                free(queue_msg);
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }
            error_size = atoi(size_buf);
            
            queue_msg->msg.error = malloc(error_size);
            /*lamvanei to onoma tou sfalmatos*/
            nbytes = read(sock, queue_msg->msg.error, error_size);
            if (nbytes < 0)
            {
                perror("read");
                free(queue_msg->msg.error);
                free(queue_msg);
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }

            printf("Received error from server: %s!\n", queue_msg->msg.error);
            free(queue_msg->msg.error);
            free(queue_msg);
            close(sock);
            exit(EXIT_FAILURE);
        }
        else
        {
            printf("Wrong message received!\n");
            free(queue_msg);
            queue_destroy(server_queue, array_size);
            queue_destroy(threads_queue, array_size);
            queue_destroy(calc_queue, array_size);
            free(threads);
            close(sock);
            exit(EXIT_FAILURE);
        }
    }

    printf("Got all messages from the server! Queue:\n");
    queue_print(server_queue);
}

void synchronize()
{
    qnode_ptr qserver, qlocal;

    qserver = server_queue->first;
    qlocal = threads_queue->first;

    while (qserver != NULL)
    {
        while (qlocal != NULL && strcmp(qlocal->msg.filename, qserver->msg.filename) < 0)
        {
            /*diagrafi arxeiou pou yparxei ston client kai oxi ston server*/
            char *fd = malloc(strlen(qlocal->msg.directory) + strlen(qlocal->msg.filename) + 1);
            strcpy(fd, qlocal->msg.directory);
            strcat(fd, qlocal->msg.filename);

            printf("Deleting file %s\n", fd);
            if (remove(fd) != 0)
                perror("Error deleting file");
            else
                printf("File successfully deleted\n");

            free(fd);

            qlocal = qlocal->next;
        }

        if (qlocal != NULL && !strcmp(qlocal->msg.filename, qserver->msg.filename))
        {                    
            /*sygrisi kai syngronismos idion arxeion*/
            comparison(qlocal, qserver, 0, qlocal->msg.filename, qserver->msg.filesize);

            qlocal = qlocal->next;
        }
        else
        {
            /*antigrafi arxeion pou den yparxoun katholou ston client*/
            int i = 1;
            /*ypologismos arithmou selidon tou arxeiou pros antigrafi*/
            int pages_num = qserver->msg.filesize / PAGE_SIZE;
            if (qserver->msg.filesize % PAGE_SIZE != 0)
                pages_num++;

            FILE * fp;

            printf("Creating file: %s with size: %d\n", qserver->msg.filename, qserver->msg.filesize);

            char *fd = malloc(strlen(directory) + strlen(qserver->msg.filename) + 1);
            strcpy(fd, directory);
            strcat(fd, qserver->msg.filename);
            if ((fp = fopen(fd, "ab")) == NULL)
            {
                perror("fopen failed");
                free(fd);
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }

            free(fd);

            while (i <= pages_num)
            {
                int nbytes, dir_size, filen_size;
                char dtmp[12], ftmp[12];
                message *msg = malloc(sizeof (message));
                msg->pageno = i;
                msg->id = GETDATA;
                
                /*stelnei ston server minima getdata*/
                nbytes = write(sock, msg, sizeof (message));
                if (nbytes < 0)
                {
                    perror("write");
                    free(msg);
                    queue_destroy(server_queue, array_size);
                    queue_destroy(threads_queue, array_size);
                    queue_destroy(calc_queue, array_size);
                    free(threads);
                    close(sock);
                    exit(EXIT_FAILURE);
                }
                free(msg);

                /*stelnei to megethos tou onomatos tou directory*/
                dir_size = strlen(directory) + 1;
                sprintf(dtmp, "%d", dir_size);
                nbytes = write(sock, dtmp, 12 * sizeof (char));
                if (nbytes < 0)
                {
                    perror("write");
                    queue_destroy(server_queue, array_size);
                    queue_destroy(threads_queue, array_size);
                    queue_destroy(calc_queue, array_size);
                    free(threads);
                    close(sock);
                    exit(EXIT_FAILURE);
                }

                /*stelnei to onoma tou directory*/
                nbytes = write(sock, directory, dir_size);
                if (nbytes < 0)
                {
                    perror("write");
                    queue_destroy(server_queue, array_size);
                    queue_destroy(threads_queue, array_size);
                    queue_destroy(calc_queue, array_size);
                    free(threads);
                    close(sock);
                    exit(EXIT_FAILURE);
                }

                /*stelnei to megethos tou onomatos tou arxeiou*/
                filen_size = strlen(qserver->msg.filename) + 1;
                sprintf(ftmp, "%d", filen_size);
                nbytes = write(sock, ftmp, 12 * sizeof (char));
                if (nbytes < 0)
                {
                    perror("write");
                    queue_destroy(server_queue, array_size);
                    queue_destroy(threads_queue, array_size);
                    queue_destroy(calc_queue, array_size);
                    free(threads);
                    close(sock);
                    exit(EXIT_FAILURE);
                }

                /*stelnei to onoma tou arxeiou*/
                nbytes = write(sock, qserver->msg.filename, filen_size);
                if (nbytes < 0)
                {
                    perror("write");
                    queue_destroy(server_queue, array_size);
                    queue_destroy(threads_queue, array_size);
                    queue_destroy(calc_queue, array_size);
                    free(threads);
                    close(sock);
                    exit(EXIT_FAILURE);
                }

                qnode_ptr queue_msg = malloc(sizeof (qnode));
                
                /*lamvanei to minima data apo ton server*/
                nbytes = read(sock, &(queue_msg->msg), sizeof (message));
                if (nbytes < 0)
                {
                    perror("read");
                    free(queue_msg);
                    queue_destroy(server_queue, array_size);
                    queue_destroy(threads_queue, array_size);
                    queue_destroy(calc_queue, array_size);
                    free(threads);
                    close(sock);
                    exit(EXIT_FAILURE);
                }

                char size_buf[12];
                int fname_size;

                /*lamvanei to megethos tou onomatos tou arxeiou*/
                nbytes = read(sock, size_buf, 12 * sizeof (char));
                if (nbytes < 0)
                {
                    perror("read");
                    free(queue_msg);
                    queue_destroy(server_queue, array_size);
                    queue_destroy(threads_queue, array_size);
                    queue_destroy(calc_queue, array_size);
                    free(threads);
                    close(sock);
                    exit(EXIT_FAILURE);
                }
                fname_size = atoi(size_buf);

                /*lamvanei to onoma tou arxeiou*/
                queue_msg->msg.filename = malloc(fname_size);
                nbytes = read(sock, queue_msg->msg.filename, fname_size);
                if (nbytes < 0)
                {
                    perror("read");
                    free(queue_msg->msg.filename);
                    free(queue_msg);
                    queue_destroy(server_queue, array_size);
                    queue_destroy(threads_queue, array_size);
                    queue_destroy(calc_queue, array_size);
                    free(threads);
                    close(sock);
                    exit(EXIT_FAILURE);
                }

                /*ypologizei to megethos tis selidas an einai i teleutaia*/
                int page_size;
                if (queue_msg->msg.pageno * PAGE_SIZE > queue_msg->msg.filesize)
                    page_size = queue_msg->msg.filesize - (queue_msg->msg.pageno - 1) * PAGE_SIZE;
                else
                    page_size = PAGE_SIZE;

                /*lamvanei tin selida pou zitithike*/
                int total_bytes = 0;
                nbytes = 0;
                queue_msg->msg.pagedata = malloc(page_size * sizeof (char));
                while (total_bytes < page_size)
                {
                    nbytes = read(sock, queue_msg->msg.pagedata + total_bytes, page_size - total_bytes);
                    if (nbytes < 0)
                    {
                        perror("read");
                        free(queue_msg->msg.pagedata);
                        free(queue_msg->msg.filename);
                        free(queue_msg);
                        queue_destroy(server_queue, array_size);
                        queue_destroy(threads_queue, array_size);
                        queue_destroy(calc_queue, array_size);
                        free(threads);
                        close(sock);
                        exit(EXIT_FAILURE);
                    }
                    total_bytes += nbytes;
                }

                /*enimeronei tin selida tou arxeio me ta nea dedomena*/
                fwrite(queue_msg->msg.pagedata, 1, page_size, fp);

                free(queue_msg->msg.filename);
                free(queue_msg->msg.pagedata);
                free(queue_msg);
                /*synexizei me tis epomenes selides*/
                i++;
            }

            printf("File created successfully!\n");
            fclose(fp);
        }
        qserver = qserver->next;
    }
}


void comparison(qnode_ptr node1, qnode_ptr node2, int position, char *filename, int filesize)
{
    int first_page = ((((int) pow(tree_fanout, tree_height)) - 1) / (tree_fanout - 1));
    int last_page = ((((int) pow(tree_fanout, tree_height + 1)) - 1) / (tree_fanout - 1)) - 1;

    /*an einai se fyllo*/
    if (position >= first_page && position <= last_page)
    {
        FILE *fp;
        
        char *fd = malloc(strlen(directory) + strlen(filename) + 1);
        strcpy(fd, directory);
        strcat(fd, filename);
        if ((fp = fopen(fd, "r+b")) == NULL)
        {
            perror("fopen failed");
            queue_destroy(server_queue, array_size);
            queue_destroy(threads_queue, array_size);
            queue_destroy(calc_queue, array_size);
            free(threads);
            close(sock);
            exit(EXIT_FAILURE);
        }
        free(fd);        
        
        /*sygrinei ta digests ton fyllon*/        
        if (memcmp(node1->msg.hash_tree[position], node2->msg.hash_tree[position], 4 * sizeof (unsigned int))
                && position - first_page <= filesize/PAGE_SIZE )
        {
            int nbytes, dir_size, filen_size;
            char dtmp[12], ftmp[12];
            message *msg = malloc(sizeof (message));
            msg->pageno = position - first_page + 1;
            /*an den einai idia ektyponei ton arihmo selidas me tin diafora*/
            printf("File %s client and server versions differ in page %d\n",
                    node1->msg.filename, msg->pageno);

            msg->id = GETDATA;
            /*stelnei ston server minima getdata*/
            nbytes = write(sock, msg, sizeof (message));
            if (nbytes < 0)
            {
                perror("write");
                free(msg);
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }
            free(msg);

            /*stelnei to megethos tou onomatos tou directory*/
            dir_size = strlen(directory) + 1;
            sprintf(dtmp, "%d", dir_size);
            nbytes = write(sock, dtmp, 12 * sizeof (char));
            if (nbytes < 0)
            {
                perror("write");
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }

            /*stelnei to onoma tou directory*/
            nbytes = write(sock, directory, dir_size);
            if (nbytes < 0)
            {
                perror("write");
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }

            /*stelnei to megethos tou onomatos tou arxeiou*/
            filen_size = strlen(filename) + 1;
            sprintf(ftmp, "%d", filen_size);
            nbytes = write(sock, ftmp, 12 * sizeof (char));
            if (nbytes < 0)
            {
                perror("write");
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }

            /*lamvanei to minima data apo ton server*/
            nbytes = write(sock, filename, filen_size);
            if (nbytes < 0)
            {
                perror("write");
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }

            qnode_ptr queue_msg = malloc(sizeof (qnode));
            /*lamvanei to minima data apo ton server*/
            nbytes = read(sock, &(queue_msg->msg), sizeof (message));
            if (nbytes < 0)
            {
                perror("read");
                free(queue_msg);
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }

            char size_buf[12];
            int fname_size;

            /*lamvanei to megethos tou onomatos tou arxeiou*/
            nbytes = read(sock, size_buf, 12 * sizeof (char));
            if (nbytes < 0)
            {
                perror("read");
                free(queue_msg);
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }
            fname_size = atoi(size_buf);

            /*lamvanei to onoma tou arxeiou*/
            queue_msg->msg.filename = malloc(fname_size);
            nbytes = read(sock, queue_msg->msg.filename, fname_size);
            if (nbytes < 0)
            {
                perror("read");
                free(queue_msg->msg.filename);
                free(queue_msg);
                queue_destroy(server_queue, array_size);
                queue_destroy(threads_queue, array_size);
                queue_destroy(calc_queue, array_size);
                free(threads);
                close(sock);
                exit(EXIT_FAILURE);
            }

            /*ypologizei to megethos tis selidas an einai i teleutaia*/
            int page_size;
            if (queue_msg->msg.pageno * PAGE_SIZE > queue_msg->msg.filesize)
                page_size = queue_msg->msg.filesize - (queue_msg->msg.pageno - 1) * PAGE_SIZE;
            else
                page_size = PAGE_SIZE;

            /*lamvanei tin selida pou zitithike*/
            int total_bytes = 0;
            nbytes = 0;
            queue_msg->msg.pagedata = malloc(page_size * sizeof (char));
            while (total_bytes < page_size)
            {
                nbytes = read(sock, queue_msg->msg.pagedata + total_bytes, page_size - total_bytes);
                if (nbytes < 0)
                {
                    perror("read");
                    free(queue_msg->msg.pagedata);
                    free(queue_msg->msg.filename);
                    free(queue_msg);
                    queue_destroy(server_queue, array_size);
                    queue_destroy(threads_queue, array_size);
                    queue_destroy(calc_queue, array_size);
                    free(threads);
                    close(sock);
                    exit(EXIT_FAILURE);
                }
                total_bytes += nbytes;
            }

            /*enimeronei tin selida tou arxeio me ta nea dedomena*/
            fseek(fp, (queue_msg->msg.pageno - 1) * PAGE_SIZE, SEEK_SET);
            fwrite(queue_msg->msg.pagedata, 1, page_size, fp);

            free(queue_msg->msg.filename);
            free(queue_msg->msg.pagedata);
            free(queue_msg);           
        }
        
        ftruncate(fileno(fp), filesize);
        fclose(fp);
    }
    else
    {
        /*sygrinei ta digests ton komvon*/
        if (!memcmp(node1->msg.hash_tree[position], node2->msg.hash_tree[position], 4 * sizeof (unsigned int)))
            /*an einai idia stamataei*/
            return;
            /*allios kalei ti comparison gia ta paidia tou komvou autou*/
        else
        {
            int cur_pos = position;
            for (position = (tree_fanout * (position + 1)) - tree_fanout + 1; position < (tree_fanout * (cur_pos + 1)) + 1; position++)
                comparison(node1, node2, position, filename, filesize);
        }
    }
}

void *worker_thread(void *argp)
{
    while (1)
    {
        pthread_mutex_lock(&queue_empty_mtx);
        while (!queue_not_empty(calc_queue) && running)
        {
            pthread_cond_wait(&queue_empty_cond, &queue_empty_mtx);
        }

        if (!running)
        {
            pthread_mutex_unlock(&queue_empty_mtx);
            break;
        }

        /*exagagei ena minima apo tin oura me ta calculata hash*/
        qnode_ptr temp_node;
        temp_node = queue_extract(calc_queue);

        if (calc_queue->size == queue_size - 1)
        {
            pthread_mutex_lock(&queue_full_mtx);
            pthread_cond_signal(&queue_full_cond);
            pthread_mutex_unlock(&queue_full_mtx);
        }
        pthread_mutex_unlock(&queue_empty_mtx);

        fflush(stdout);

        /*to gemizei me tis aparaitites plirofories*/
        temp_node->msg.id = FILEINFO;
        struct stat st;
        char *fd = malloc(strlen(directory) + strlen(temp_node->msg.filename) + 1);
        strcpy(fd, directory);
        strcat(fd, temp_node->msg.filename);
        stat(fd, &st);
        temp_node->msg.filesize = st.st_size;

        /*ypologizei to hash tree*/
        if (hash_tree_create(&temp_node->msg.hash_tree, fd, tree_height, tree_fanout))
        {
            free(fd);
            queue_destroy(server_queue, array_size);
            queue_destroy(threads_queue, array_size);
            queue_destroy(calc_queue, array_size);
            free(threads);
            close(sock);            
            exit(EXIT_FAILURE);
        }

        free(fd);

        /*to eisagei stin oura me ta apotelesmata ton threrads*/
        pthread_mutex_lock(&tqueue_ins_mtx);
        queue_insert(threads_queue, temp_node);
        pthread_mutex_unlock(&tqueue_ins_mtx);

        pthread_mutex_lock(&queue_empty_mtx);
        if (calc_queue->size == 0)
        {
            pthread_cond_signal(&queue_empty_cond);
        }
        pthread_mutex_unlock(&queue_empty_mtx);

        if (!running)
        {
            break;
        }
    }

    pthread_exit(NULL);
}
