/* File: Hash_tree.c */

#include "Hash_tree.h"

int hash_tree_create(unsigned int ***tree, char *file, int tree_height,  int tree_fanout)
{
    int i, j, k;

    int array_size = ((((int) pow(tree_fanout, tree_height + 1)) - 1) / (tree_fanout - 1));    
    if ((*tree =  malloc(array_size * sizeof (unsigned int*))) == NULL)
    {
        perror("malloc failed");
        return 1;
    }
    for(i = 0; i < array_size; i++)
    {
        if (((*tree)[i] = malloc(4 * sizeof (unsigned int))) == NULL)
        {
            perror("malloc failed");
            free(*tree);
            return 1;
        }
    }
    
    FILE *fp;
    size_t read;
    char* buffer;
    if ((buffer = malloc(PAGE_SIZE * sizeof (char))) == NULL)
    {
        perror("malloc failed");
        for(i = 0; i < array_size; i++)
        {
            free((*tree)[i]);
        }          
        free(*tree);
        return 1;
    }             

    fp = fopen(file, "rb");
    if (fp == NULL)
    {
        perror("Could not open file");
        for(i = 0; i < array_size; i++)
        {
            free((*tree)[i]);
        }          
        free(*tree);        
        free(buffer);
        return 1;
    }

    /*ypologismos ton digests ton selidon gia to arxeio*/    
    i = 1;
    MD5_CTX context;
    int page_counter = 0;
    int first_page = ((((int) pow(tree_fanout, tree_height)) - 1)/(tree_fanout -1));
    do
    { 
        read = fread(buffer, 1, PAGE_SIZE, fp);        
        if (read > 0)
        {
            
            if (read < PAGE_SIZE)
            {
                memset(buffer + read, 0, (PAGE_SIZE - read));
            }            
            
            /*ypologismos tis sinopsis (digest)*/
            MD5Init(&context);
            MD5Update(&context, buffer, strlen(buffer));           
            MD5Final((*tree)[first_page + page_counter], &context);             
        }
        else
        {
            if(i == 1)
            {
                printf("Empty file!\n");
                return 0;
            }
        }
        i++;
        page_counter++;
    }while (read == PAGE_SIZE);         
    
    /*gemisma kenon selidon*/          
    while(first_page + page_counter < array_size) 
    {              
        /*eisagogi tis synopsis sto dentro*/
        for(j = 0; j < 4; j++)
            (*tree)[first_page + page_counter][j] = 0;    
        
        page_counter++;        
    }    

    /*gemisma olon ton komvon*/
    i = 1;  
    int first_ofl;
    int last_ofl;
    while(i <= tree_height)
    {        
        first_ofl = ((((int) pow(tree_fanout, tree_height-i)) - 1)/(tree_fanout -1))+1;
        last_ofl = ((((int) pow(tree_fanout, tree_height-i+1)) - 1)/(tree_fanout -1));    
        for(j = first_ofl; j <= last_ofl; j++)
        {
            MD5Init( &context );
            for(k = (tree_fanout*j) - tree_fanout + 1; k <= (tree_fanout*j); k++)
            {
                MD5Update( &context , (char *)(*tree)[k] , 4*sizeof(unsigned int));
            }
            MD5Final(  (*tree)[j-1], &context );
        }
        i++;        
    }
    
    if (ferror(fp))
        perror("An error occured during reading");
    
    fclose(fp);    
    
    free(buffer);  
    buffer = NULL;
    
    return 0;    
}
