/* File: filesynd.c */


#include "filesynd_functions.h"

int ext_port, inter_port, queue_size, pool_size, tree_height, tree_fanout, running = 1;
pthread_t *threads;
qhead_ptr threads_queue;
pthread_mutex_t queue_empty_mtx = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t queue_empty_cond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t queue_full_mtx = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t queue_full_cond = PTHREAD_COND_INITIALIZER;

int main(int argc, char** argv)
{
    /*orismos tou action pou tha kaleitai otan erthei ena apo ta parakato simata*/
    act.sa_handler = interrupt_handler;
    sigfillset(&(act.sa_mask));
    sigaction(SIGINT, &act, NULL);
    sigaction(SIGTERM, &act, NULL);
    sigaction(SIGHUP, &act, NULL);
    /*an erthei sima SIGPIPE to agnoei kai diaxeirizetai to sfalma stis read kai write*/
    signal(SIGPIPE, SIG_IGN);

    /*elengxos tou arithmou ton arguments*/
    if (argc < 6)
    {
        printf("Not enough arguments! The program will now exit...\n");
        return 0;
    }

    /*perasma tous se metavlites*/
    ext_port = atoi(argv[1]);
    if (ext_port < 0 || ext_port > 65535)
    {
        printf("Port number must be within 0-65535 range! Try again..\n");
    }
    pool_size = atoi(argv[2]);
    queue_size = atoi(argv[3]);
    tree_height = atoi(argv[4]);
    tree_fanout = atoi(argv[5]);

    int i, ext_sock, new_sock, inter_sock;
    struct sockaddr_in client_addr, server_addr;
    fd_set master_set, active_set;

    /*Dimiourgeitai to external socket sto port apo ta args*/
    if (sock_create(&ext_sock, ext_port, 10))
        return 1;

    /*Dimiourgeitai to internal socket se ena tyxaio diathesimo port*/
    if (sock_create(&inter_sock, 0, pool_size))
    {
        close(ext_sock);
        return 1;
    }

    /*Vriskei auto to port kai to apothikeuei*/
    int size = sizeof (server_addr);
    new_sock = getsockname(inter_sock, (struct sockaddr *) &server_addr, &size);
    inter_port = ntohs(server_addr.sin_port);

    /*Arxikopoiiei to set ton sockets*/
    FD_ZERO(&master_set);
    FD_SET(ext_sock, &master_set);
    FD_SET(inter_sock, &master_set);

    /*Dimiourgia ouras minimaton*/
    if (queue_init(&threads_queue))
    {
        close(ext_sock);
        close(inter_sock);
        free(threads);
        return 1;
    }

    /*Dimiourgia tou threads pool*/
    if ((threads = malloc(pool_size * sizeof (pthread_t))) == NULL)
    {
        perror("malloc failed");
        close(ext_sock);
        close(inter_sock);
        return 1;
    }
    for (i = 0; i < pool_size; i++)
    {
        if (pthread_create(&threads[i], NULL, worker_thread, NULL))
        {
            perror("pthread_create failed");
            close(ext_sock);
            close(inter_sock);
            free(threads);
            return 1;
        }

    }

    while (1)
    {
        active_set = master_set;

        /*Mplokarei mexri na katafthasei mia i perissoteres syndeseis*/
        if (select(FD_SETSIZE, &active_set, NULL, NULL, NULL) < 0)
        {
            perror("select failed");
            close(ext_sock);
            close(inter_sock);
            free(threads);
            return 1;
        }

        /*Eksipiretounte oles oi syndeseis me energa aitimata*/
        for (i = 0; i < FD_SETSIZE; ++i)
            if (FD_ISSET(i, &active_set))
            {
                /*an irthe aitima gia syndesi sto eksoteriko listening socket*/
                if (i == ext_sock)
                {
                    printf("Listening %d socket received connection!\n", i);
                    new_sock = accept(ext_sock, (struct sockaddr *) &client_addr, &size);
                    if (new_sock < 0)
                    {
                        perror("accept failed");
                        continue;
                    }
                    
                    char addr[INET_ADDRSTRLEN];
                    inet_ntop(AF_INET, &(client_addr.sin_addr), addr, INET_ADDRSTRLEN);
                    printf("Accepted connection in socket %d from client %s!\n", new_sock, addr);
                    
                    FD_SET(new_sock, &master_set);
                }
                /*an irthe aitima gia syndesi sto esoteriko listening socket*/
                else if (i == inter_sock)
                {
                    printf("Listening %d socket received connection!\n", i);
                    new_sock = accept(inter_sock, (struct sockaddr *) &client_addr, &size);

                    if (new_sock < 0)
                    {
                        perror("accept failed");
                        continue;
                    }
                    FD_SET(new_sock, &master_set);
                }
                /*an irthe minima idi syndedemenou pelati i worker thread*/
                else
                {
                    int bytes_num;
                    qnode_ptr queue_msg = malloc(sizeof (qnode));

                    /*lamvanei to eiserxomeno minima*/
                    bytes_num = read(i, &(queue_msg->msg), sizeof (message));
                    if (bytes_num < 0)
                    {
                        perror("read failed");
                        free(queue_msg);
                        FD_CLR(i, &master_set);  
                        close(i);
                        continue;
                    }
                    else if(bytes_num == 0)
                    {
                        free(queue_msg);
                        FD_CLR(i, &master_set);  
                        close(i);
                        continue;
                    }

                    /*analoga me ton typo tou minimatos ektelei tis aparaitites energeies*/
                    switch (queue_msg->msg.id)
                    {
                        case GETFILES:
                        {
                            if(getfiles(queue_msg, i))
                                FD_CLR(i, &master_set);

                            break;
                        }
                        case EXPECT:
                        {
                            if(expect(queue_msg, i))
                                FD_CLR(i, &master_set);

                            break;
                        }
                        case FILEINFO:
                        {
                            if(fileinfo(queue_msg, i))
                                FD_CLR(i, &master_set);                                                

                            break;
                        }
                        case GETDATA:
                        {
                            if(getdata(queue_msg, i))
                                FD_CLR(i, &master_set);                               

                            break;
                        }
                        case DATA:
                        {
                            if(data(queue_msg, i))
                                FD_CLR(i, &master_set);                             
                            
                            break;
                        }
                        case ERROR:
                        {
                            if(error(queue_msg, i))
                                FD_CLR(i, &master_set);                               

                            break;                            
                        }
                        default:
                        {
                            printf("Wrong message received!\n");
                            raise(SIGINT);
                        }
                    }
                }
            }
    }

    return 0;
}
