/* File: filesynd_functions.c */


#include "filesynd_functions.h"

/*mutex gia tin klisi ton opendir, readdir, closedir apo ta threads*/
pthread_mutex_t dir_mtx = PTHREAD_MUTEX_INITIALIZER;


int sock_create(int *sock, int port_num, int listen_num)
{
    int opt = 1;
    struct sockaddr_in server_addr;

    if ((*sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        perror("socket failed");
        return 1;
    }

    /*Epitrepetai i epanaxrisimopoiisi tou socket descriptor*/
    if (setsockopt(*sock, SOL_SOCKET, SO_REUSEADDR, (char *) &opt, sizeof (opt)) < 0)
    {
        perror("setsockopt failed");
        close(*sock);
        return 1;
    }

    /*Apenergopoiisi tou Nagle buffering algorithmou*/
    if (setsockopt(*sock, IPPROTO_TCP, TCP_NODELAY, (char *) &opt, sizeof (opt)) < 0)
    {
        perror("setsockopt failed");
        close(*sock);
        return 1;
    }

    /*Ginetai to bind me to socket*/
    server_addr.sin_family = AF_INET;
    server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    server_addr.sin_port = htons(port_num);
    if (bind(*sock, (struct sockaddr *) &server_addr, sizeof (server_addr)) < 0)
    {
        perror("bind failed");
        close(*sock);
        return 1;
    }

    /*Ginetai to listening gia tis eiserxomenes syndeseis*/
    if (listen(*sock, listen_num) < 0)
    {
        perror("listen failed");
        close(*sock);
        return 1;
    }
    
    return 0;
}


int getfiles(qnode_ptr queue_msg, int socket)
{
    char size_buf[12];
    int dir_size, bytes_num;

    /*lamvanei to mikos tou onomatos tou directory*/
    bytes_num = read(socket, size_buf, 12 * sizeof (char));
    if (bytes_num < 0)
    {
        perror("read failed");
        free(queue_msg);
        close(socket);
        return 1;
    }
    dir_size = atoi(size_buf);

    /*lamvanei to onoma tou directory*/
    queue_msg->msg.directory = malloc(dir_size);
    bytes_num = read(socket, queue_msg->msg.directory, dir_size);
    if (bytes_num < 0)
    {
        perror("read failed");
        free(queue_msg->msg.directory);
        free(queue_msg);
        close(socket);
        return 1;
    }
    else if (bytes_num == 0)
    {
        printf("Directory name not received. Closing connection %d", socket);
        free(queue_msg->msg.directory);
        free(queue_msg);
        close(socket);
        return 1;
    }
    else
    {
        printf("Server got directory name: `%s' with size of %d\n", queue_msg->msg.directory, dir_size);
        /*emploutismos minimatos me tonn arithmo tou socket tou client*/
        queue_msg->socket = socket;

        pthread_mutex_lock(&queue_empty_mtx);
        while (threads_queue->size == queue_size && running)
        {
            pthread_mutex_unlock(&queue_empty_mtx);
            pthread_mutex_lock(&queue_full_mtx);
            pthread_cond_wait(&queue_full_cond, &queue_full_mtx);
            pthread_mutex_unlock(&queue_full_mtx);
            pthread_mutex_lock(&queue_empty_mtx);
        }

        /*totpothetisi minimatos stin oura*/
        queue_insert(threads_queue, queue_msg);

        pthread_mutex_unlock(&queue_empty_mtx);
        pthread_cond_signal(&queue_empty_cond);
    }
    
    return 0;
}


int expect(qnode_ptr queue_msg, int socket)
{
    int client_sock, bytes_num;

    /*lamvanei prota gia poio socket pelati prokeitai*/
    bytes_num = read(socket, &client_sock, sizeof (int));
    if (bytes_num < 0)
    {
        perror("read failed");
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*stelnei ston pelati to minima me to arithmo ton arxeion pou tha perimenei*/
    bytes_num = write(client_sock, &(queue_msg->msg), sizeof (message));
    if (bytes_num < 0)
    {
        perror("write failed");
        free(queue_msg);
        close(socket);
        return 1;
    }

    free(queue_msg); 
    return 0;
}

int fileinfo(qnode_ptr queue_msg, int socket)
{
    int client_sock, bytes_num, fname_size, k;
    char size_buf[12];
    int array_size = ((((int) pow(tree_fanout, tree_height + 1)) - 1) / (tree_fanout - 1));

    /*lamvanei prota gia poio socket pelati prokeitai*/
    bytes_num = read(socket, &client_sock, sizeof (int));
    if (bytes_num < 0)
    {
        perror("read failed");
        free(queue_msg->msg.directory);
        free(queue_msg->msg.filename);
        for (k = 0; k < array_size; k++)
        {
            free(queue_msg->msg.hash_tree[k]);
        }
        free(queue_msg->msg.hash_tree);
        free(queue_msg);        
        close(socket);
        return 1;
    }

    /*stelnei ston pelati to minima fileinfo*/
    bytes_num = write(client_sock, &(queue_msg->msg), sizeof (message));
    if (bytes_num < 0)
    {
        perror("write failed");
        free(queue_msg->msg.directory);
        free(queue_msg->msg.filename);
        for (k = 0; k < array_size; k++)
        {
            free(queue_msg->msg.hash_tree[k]);
        }
        free(queue_msg->msg.hash_tree);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*stelnei to megethos tou onomatos tou arxeiou*/
    fname_size = strlen(queue_msg->msg.filename) + 1;
    sprintf(size_buf, "%d", fname_size);
    bytes_num = write(client_sock, size_buf, 12 * sizeof (char));
    if (bytes_num < 0)
    {
        perror("write failed");
        free(queue_msg->msg.directory);
        free(queue_msg->msg.filename);
        for (k = 0; k < array_size; k++)
        {
            free(queue_msg->msg.hash_tree[k]);
        }
        free(queue_msg->msg.hash_tree);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*stelnei to onoma tou arxeiou*/
    bytes_num = write(client_sock, queue_msg->msg.filename, fname_size);
    if (bytes_num < 0)
    {
        perror("write failed");
        free(queue_msg->msg.directory);
        free(queue_msg->msg.filename);
        for (k = 0; k < array_size; k++)
        {
            free(queue_msg->msg.hash_tree[k]);
        }
        free(queue_msg->msg.hash_tree);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*etoimazei to block me tin pliroforia gia to hash tree tou arxeiou*/
    int j;
    int nodes_num = ((((int) pow(tree_fanout, tree_height)) - 1) / (tree_fanout - 1));
    int payload = queue_msg->msg.filesize / PAGE_SIZE;
    if (queue_msg->msg.filesize % PAGE_SIZE != 0)
    {
        payload += 1;
    }
    char *tree_buf = malloc((nodes_num + payload) * 4 * sizeof (unsigned int));
    for (j = 0; j < nodes_num + payload; j++)
    {
        memcpy(tree_buf + j * 4 * sizeof (unsigned int), queue_msg->msg.hash_tree[j], 4 * sizeof (unsigned int));
    }

    /*stelnei to block me tin pliroforia gia to hash tree*/
    bytes_num = write(client_sock, tree_buf, (nodes_num + payload) * 4 * sizeof (unsigned int));
    if (bytes_num < 0)
    {
        perror("3.1 write failed");
        free(tree_buf);
        free(queue_msg->msg.directory);
        free(queue_msg->msg.filename);
        for (k = 0; k < array_size; k++)
        {
            free(queue_msg->msg.hash_tree[k]);
        }
        free(queue_msg->msg.hash_tree);
        free(queue_msg);
        close(socket);
        return 1;
    }

    printf("Send message %d filename: %s size: %d\n", (int) queue_msg->msg.id, queue_msg->msg.filename, queue_msg->msg.filesize);

    free(tree_buf);
    free(queue_msg->msg.directory);
    free(queue_msg->msg.filename);
    for (k = 0; k < array_size; k++)
    {
        free(queue_msg->msg.hash_tree[k]);
    }
    free(queue_msg->msg.hash_tree);
    free(queue_msg);
    
    return 0;
}


int getdata(qnode_ptr queue_msg, int socket)
{
    char dsize_buf[12], fsize_buf[12], bytes_num;
    int dir_size, filen_size;

    /*lamvanei to mikos tou onomatos tou directory*/
    bytes_num = read(socket, dsize_buf, 12 * sizeof (char));
    if (bytes_num < 0)
    {
        perror("read failed");
        free(queue_msg);
        close(socket);
        return 1;
    }
    dir_size = atoi(dsize_buf);

    /*lamvanei to onoma tou directory*/
    queue_msg->msg.directory = malloc(dir_size);
    bytes_num = read(socket, queue_msg->msg.directory, dir_size);
    if (bytes_num < 0)
    {
        perror("read failed");
        free(queue_msg->msg.directory);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*lamvanei to mikos tou onomatos tou arxeiou*/
    bytes_num = read(socket, fsize_buf, 12 * sizeof (char));
    if (bytes_num < 0)
    {
        perror("read failed");
        free(queue_msg->msg.directory);
        free(queue_msg);
        close(socket);
        return 1;
    }
    filen_size = atoi(fsize_buf);

    /*lamvanei to onoma tou arxeiou*/
    queue_msg->msg.filename = malloc(filen_size);
    bytes_num = read(socket, queue_msg->msg.filename, filen_size);
    if (bytes_num < 0)
    {
        perror("read failed");
        free(queue_msg->msg.filename);
        free(queue_msg->msg.directory);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*emploutismos minimatos me ton arithmo tou socket tou client*/
    queue_msg->socket = socket;

    pthread_mutex_lock(&queue_empty_mtx);
    while (threads_queue->size == queue_size && running)
    {
        pthread_mutex_unlock(&queue_empty_mtx);
        pthread_mutex_lock(&queue_full_mtx);
        pthread_cond_wait(&queue_full_cond, &queue_full_mtx);
        pthread_mutex_unlock(&queue_full_mtx);
        pthread_mutex_lock(&queue_empty_mtx);
    }

    /*topothetisi minimatos stin oura*/
    queue_insert(threads_queue, queue_msg);

    pthread_mutex_unlock(&queue_empty_mtx);
    pthread_cond_signal(&queue_empty_cond);   
    
    return 0;
}


int data(qnode_ptr queue_msg, int socket)
{
    int client_sock, bytes_num, fname_size;
    char size_buf[12];

    /*lamvanei prota gia poio socket pelati prokeitai*/
    bytes_num = read(socket, &client_sock, sizeof (int));
    if (bytes_num < 0)
    {
        perror("read failed");
        free(queue_msg->msg.directory);
        free(queue_msg->msg.filename);
        free(queue_msg->msg.pagedata);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*stelnei ston pelati to minima data*/
    bytes_num = write(client_sock, &(queue_msg->msg), sizeof (message));
    if (bytes_num < 0)
    {
        perror("write failed");
        free(queue_msg->msg.directory);
        free(queue_msg->msg.filename);
        free(queue_msg->msg.pagedata);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*stelnei to megethos tou onomatos tou arxeiou*/
    fname_size = strlen(queue_msg->msg.filename) + 1;
    sprintf(size_buf, "%d", fname_size);
    bytes_num = write(client_sock, size_buf, 12 * sizeof (char));
    if (bytes_num < 0)
    {
        perror("write failed");
        free(queue_msg->msg.directory);
        free(queue_msg->msg.filename);
        free(queue_msg->msg.pagedata);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*stelnei to onoma tou arxeiou*/
    bytes_num = write(client_sock, queue_msg->msg.filename, fname_size);
    if (bytes_num < 0)
    {
        perror("write failed");
        free(queue_msg->msg.directory);
        free(queue_msg->msg.filename);
        free(queue_msg->msg.pagedata);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*ypologizei to megethos tis selidas an einai i teleutaia*/
    int page_size;
    if (queue_msg->msg.pageno * PAGE_SIZE > queue_msg->msg.filesize)
        page_size = queue_msg->msg.filesize - (queue_msg->msg.pageno - 1) * PAGE_SIZE;
    else
        page_size = PAGE_SIZE;

    /*stelnei tin selida pou zitithike*/
    bytes_num = write(client_sock, queue_msg->msg.pagedata, page_size);
    if (bytes_num < 0)
    {
        perror("write failed");
        free(queue_msg->msg.directory);
        free(queue_msg->msg.filename);
        free(queue_msg->msg.pagedata);
        free(queue_msg);
        close(socket);
        return 1;
    }

    free(queue_msg->msg.directory);
    free(queue_msg->msg.filename);
    free(queue_msg->msg.pagedata);
    free(queue_msg);

    printf("Sent page data to client!\n");  
    
    return 0;
}


int error(qnode_ptr queue_msg, int socket)
{
    int client_sock, bytes_num, error_size;
    char size_buf[12];

    /*lamvanei prota gia poio socket pelati prokeitai*/
    bytes_num = read(socket, &client_sock, sizeof (int));
    if (bytes_num < 0)
    {
        perror("read failed");
        free(queue_msg->msg.error);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*stelnei ston pelati to minima error*/
    bytes_num = write(client_sock, &(queue_msg->msg), sizeof (message));
    if (bytes_num < 0)
    {
        perror("write failed");
        free(queue_msg->msg.error);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*stelnei to megethos tou onomatos sfalmatos*/
    error_size = strlen(queue_msg->msg.error) + 1;
    sprintf(size_buf, "%d", error_size);
    bytes_num = write(client_sock, size_buf, 12 * sizeof (char));
    if (bytes_num < 0)
    {
        perror("write failed");
        free(queue_msg->msg.error);
        free(queue_msg);
        close(socket);
        return 1;
    }

    /*stelnei to  onoma tou sfalmatos*/
    bytes_num = write(client_sock, queue_msg->msg.error, error_size);
    if (bytes_num < 0)
    {
        perror("write failed");
        free(queue_msg->msg.error);
        free(queue_msg);
        close(socket);
        return 1;
    }

    free(queue_msg->msg.error);
    free(queue_msg);    
    
    return 0;
}


void *worker_thread(void *argp)
{
    int sock;
    struct sockaddr_in servername;

    /*Dimiourgei to socket gia tin endoepikoinonia*/
    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        perror("socket failed");
        pthread_exit(NULL);
    }

    memset(&servername, 0, sizeof (servername));
    servername.sin_family = AF_INET;
    servername.sin_addr.s_addr = htonl(INADDR_ANY);
    servername.sin_port = htons(inter_port);

    /*syndeetai sto internal socket*/
    if (connect(sock, (struct sockaddr *) &servername, sizeof (servername)) < 0)
    {
        perror("connect failed");
        close(sock);
        pthread_exit(NULL);
    }

    printf("Thread with id %u connected!\n", (unsigned int) pthread_self());

    while (1)
    {
        pthread_mutex_lock(&queue_empty_mtx);
        while (!queue_not_empty(threads_queue) && running)
        {
            pthread_cond_wait(&queue_empty_cond, &queue_empty_mtx);
        }

        if (!running)
        {
            pthread_mutex_unlock(&queue_empty_mtx);
            break;
        }

        /*exagagei ena minima apo tin oura*/
        qnode_ptr temp_node;
        temp_node = queue_extract(threads_queue);

        if (threads_queue->size == queue_size - 1)
        {
            pthread_mutex_lock(&queue_full_mtx);
            pthread_cond_signal(&queue_full_cond);
            pthread_mutex_unlock(&queue_full_mtx);
        }

        pthread_mutex_unlock(&queue_empty_mtx);

        /*analoga me ton typo tou minimatos ektelei tis aparaitites energeies*/
        switch (temp_node->msg.id)
        {
            case GETFILES:
            {
                temp_node->msg.id = EXPECT;

                pthread_mutex_lock(&dir_mtx);

                DIR *dir_ptr;
                struct dirent *dirent_ptr;
                
                if ((dir_ptr = opendir(temp_node->msg.directory)) == NULL)
                {
                    perror("opendir failed");

                    temp_node->msg.id = ERROR;

                    temp_node->msg.error = malloc(sizeof ("Requested directory doesn't exist on server"));
                    strcpy(temp_node->msg.error, "Requested directory doesn't exist on server");
                }
                else
                {
                    temp_node->msg.expect_num = 0;

                    /*vriskei ta diathesima arxeia tou fakelou*/
                    while ((dirent_ptr = readdir(dir_ptr)) != NULL)
                    {
                        if (strcmp(dirent_ptr->d_name, ".") && strcmp(dirent_ptr->d_name, ".."))
                        {
                            /*gia kathe ena apo auta dimiourgei ena processfile minima*/
                            qnode_ptr pf_node = malloc(sizeof (qnode));
                            pf_node->socket = temp_node->socket;
                            pf_node->msg.id = PROCESSFILE;
                            pf_node->msg.directory = malloc(strlen(temp_node->msg.directory) + 1);
                            strcpy(pf_node->msg.directory, temp_node->msg.directory);
                            pf_node->msg.filename = malloc(strlen(dirent_ptr->d_name) + 1);
                            strcpy(pf_node->msg.filename, dirent_ptr->d_name);

                            pthread_mutex_lock(&queue_empty_mtx);
                            while (threads_queue->size == queue_size && running)
                            {
                                pthread_mutex_unlock(&queue_empty_mtx);
                                pthread_mutex_lock(&queue_full_mtx);
                                pthread_cond_wait(&queue_full_cond, &queue_full_mtx);
                                pthread_mutex_unlock(&queue_full_mtx);
                                pthread_mutex_lock(&queue_empty_mtx);
                            }

                            if (!running)
                            {
                                pthread_mutex_unlock(&queue_empty_mtx);
                                close(sock);
                                break;
                            }

                            /*to eisagei stin oura*/
                            queue_insert(threads_queue, pf_node);

                            pthread_mutex_unlock(&queue_empty_mtx);
                            pthread_cond_signal(&queue_empty_cond);

                            temp_node->msg.expect_num++;
                        }
                    }
                    closedir(dir_ptr);
                }

                pthread_mutex_unlock(&dir_mtx);

                int nbytes;

                /*stelnei sto main thread to minima*/
                nbytes = write(sock, &(temp_node->msg), sizeof (message));
                if (nbytes < 0)
                {
                    perror("write");
                    exit(EXIT_FAILURE);
                }

                /*stelnei sto main thread to socket gia to opoio prokeitai*/
                nbytes = write(sock, &(temp_node->socket), sizeof (int));
                if (nbytes < 0)
                {
                    perror("write");
                    exit(EXIT_FAILURE);
                }

                free(temp_node->msg.directory);

                break;
            }
            case PROCESSFILE:
            {
                /*allazei to minima se minima fileinfo*/
                temp_node->msg.id = FILEINFO;

                /*to gemizei me tis aparaitites plirofories*/
                struct stat st;
                char *fd = malloc(strlen(temp_node->msg.directory) + strlen(temp_node->msg.filename) + 1);
                strcpy(fd, temp_node->msg.directory);
                strcat(fd, temp_node->msg.filename);
                stat(fd, &st);
                temp_node->msg.filesize = st.st_size;

                /*ypologizei to hash tree*/
                if (hash_tree_create(&temp_node->msg.hash_tree, fd, tree_height, tree_fanout))
                {                   
                    temp_node->msg.id = ERROR;
                    
                    free(temp_node->msg.directory);
                    free(temp_node->msg.filename);
                    free(temp_node);                    

                    temp_node->msg.error = malloc(sizeof ("Failed to create hash tree"));
                    strcpy(temp_node->msg.error, "Failed to create hash tree");
                }

                int nbytes;

                /*stelnei sto main thread to minima*/
                nbytes = write(sock, &temp_node->msg, sizeof (message));
                if (nbytes < 0)
                {
                    perror("write");
                    exit(EXIT_FAILURE);
                }

                /*stelnei sto main thread to socket gia to opoio prokeitai*/
                nbytes = write(sock, &temp_node->socket, sizeof (int));
                if (nbytes < 0)
                {
                    perror("write");
                    exit(EXIT_FAILURE);
                }

                free(fd);

                break;
            }
            case GETDATA:
            {
                /*allazei to minima se minima data*/
                temp_node->msg.id = DATA;

                /*to gemizei me tis aparaitites plirofories*/
                FILE * fp;
                int page_size;
                struct stat st;
                char *fd = malloc(strlen(temp_node->msg.directory) + strlen(temp_node->msg.filename) + 1);
                strcpy(fd, temp_node->msg.directory);
                strcat(fd, temp_node->msg.filename);
                stat(fd, &st);
                temp_node->msg.filesize = st.st_size;

                int err = 0;
                if ((fp = fopen(fd, "rb")) == NULL)
                {
                    perror("fopen failed");
                    err = 1;
                    temp_node->msg.id = ERROR;

                    temp_node->msg.error = malloc(sizeof ("Requested file doesn't exist on server"));
                    strcpy(temp_node->msg.error, "Requested file doesn't exist on server");
                }

                if(!err)
                {
                    /*ypologizei to megethos tis seldas an einai i teleutaia*/
                    if (temp_node->msg.pageno * PAGE_SIZE > temp_node->msg.filesize)
                        page_size = temp_node->msg.filesize - (temp_node->msg.pageno - 1) * PAGE_SIZE;
                    else
                        page_size = PAGE_SIZE;

                    /*vriskei ta data tis selidas pou prpei na staloun*/
                    fseek(fp, (temp_node->msg.pageno - 1) * PAGE_SIZE, SEEK_SET);
                    temp_node->msg.pagedata = malloc(page_size * sizeof (char));
                    fread(temp_node->msg.pagedata, 1, page_size, fp);

                    fclose(fp);
                }

                int nbytes;

                /*stelnei sto main thread to minima*/                
                nbytes = write(sock, &temp_node->msg, sizeof (message));
                if (nbytes < 0)
                {
                    perror("write");
                    exit(EXIT_FAILURE);
                }

                /*stelnei sto main thread to socket gia to opoio prokeitai*/
                nbytes = write(sock, &temp_node->socket, sizeof (int));
                if (nbytes < 0)
                {
                    perror("write");
                    exit(EXIT_FAILURE);
                }

                free(fd);

                break;
            }
            default:
                printf("Wrong message!\n");
        }

        if (!running)
        {
            break;
        }
    }

    printf("Thread with id %u exiting!\n", (unsigned int) pthread_self());

    close(sock);
    pthread_exit(NULL);
}


void interrupt_handler(int signo)
{
    /*elave minima gia termatismo o server*/
    int i;
    running = 0;/*allagi tis running metavlitis*/
    printf("Shutting-down Server...\n");

    /*stelnei sima se ola ta threads na ksemplokaroun*/
    for (i = 0; i < pool_size; i++)
    {
        pthread_mutex_lock(&queue_empty_mtx);
        pthread_cond_signal(&queue_empty_cond);
        pthread_mutex_unlock(&queue_empty_mtx);
    }    
    for (i = 0; i < pool_size + 1; i++)
    {
        pthread_mutex_lock(&queue_full_mtx);
        pthread_cond_signal(&queue_full_cond);
        pthread_mutex_unlock(&queue_full_mtx);
    }

    /*ta perimenei na teleiosoun*/
    for (i = 0; i < pool_size; i++)
    {
        pthread_join(threads[i], NULL);
    }

    free(threads);
    free(threads_queue);

    close(ext_port);
    close(inter_port);

    /*epanaferei tin default symperifora ton simaton kai stelnei to sima ston eauto tou*/
    act.sa_handler = SIG_DFL;
    sigaction(signo, &act, NULL);
    raise(signo);
}
